#include <linux/sched.h>

#include <asm/dma.h>
#include <asm/io.h>
#include <asm/hardware.h>

typedef struct {
    unsigned long addr;
    unsigned long count;
    int active	:1;
    int invalid	:1;
} dma_t;

dma_t dma[6];

void arch_set_dma_mode (unsigned int dmanr, char mode)
{
    unsigned int io_ctrl, io_base, ctrl;

    switch (dmanr) {
    case 0:
	io_ctrl = ioaddr (IOMD_IO0CR) - ioaddr (IOMD_BASE);
	break;
    case 1:
	io_ctrl = ioaddr (IOMD_IO1CR) - ioaddr (IOMD_BASE);
	break;
    case 2:
	io_ctrl = ioaddr (IOMD_IO2CR) - ioaddr (IOMD_BASE);
	break;
    case 3:
	io_ctrl = ioaddr (IOMD_IO3CR) - ioaddr (IOMD_BASE);
	break;
    }
    io_base = ioaddr (IOMD_BASE);
    ctrl = arm_inb (io_base + io_ctrl);
    ctrl = mode == DMA_MODE_WRITE ? ctrl & ~DMA_CR_D : ctrl | DMA_CR_D;
    arm_outb (ctrl, io_base + io_ctrl);
}

void arch_set_dma_addr (unsigned int dmanr, unsigned long physaddr)
{
    dma[dmanr].addr = physaddr;
    dma[dmanr].invalid = 1;
}

void arch_set_dma_count (unsigned int dmanr, unsigned long count)
{
    dma[dmanr].count = count;
    dma[dmanr].invalid = 1;
}

int arch_request_dma (unsigned int dmanr)
{
    unsigned long flags;
    int ret;

    dma[dmanr].invalid = 1;
    dma[dmanr].active = 0;

    save_flags_cli (flags);

    ret = request_irq (dmanr + IRQ_DMA0, arch_dma_handle, 0, "dma", &dma[dmanr]);
    if (!ret)
	disable_irq (dmanr + IRQ_DMA0);

    restore_flags (flags);

    return ret;
}

void arch_free_dma (unsigned int dmanr)
{
    free_irq (dmanr + IRQ_DMA0, &dma[dmanr]);
    dma[dmanr].active = 0;
}

void arch_setup_dma (unsigned int dmanr)
{
    unsigned long thissize, end = 0;

    thissize = PAGE_SIZE - (dma[dmanr].addr & (PAGE_SIZE - 1));

    if (thissize > dma[dmanr].count) {
	thissize = dma[dmanr].count;
	end |= DMA_END_S;
    }

    if (thissize < dma[dmanr].transfersize)
	end |= DMA_END_L;

    end = (dma[dmanr].addr & (PAGE_SIZE - 1)) + thissize;

    /* write buff a current
    arm_outl (dma[dmanr].addr, iochbase + IOMD_IO0CURA - IOMD_IO0CURA);
     * write buff a end
     * set stop/last_xfer
    arm_outl (thisend, iochbase + IOMD_IO0ENDA - IOMD_IO0CURA);
     * update pointers
     * if more {
     *   write buff b current
     *   write buff b end
     * }
     */
}

void arch_enable_dma (unsigned int dmanr)
{
    if (!dma[dmanr].active) {
	unsigned int io_ctrl, io_base;

    	dma[dmanr].active = 1;

	switch (dmanr) {
	case 0:
	    io_ctrl = ioaddr (IOMD_IO0CR) - ioaddr (IOMD_BASE);
	    break;
	case 1:
	    io_ctrl = ioaddr (IOMD_IO1CR) - ioaddr (IOMD_BASE);
	    break;
	case 2:
	    io_ctrl = ioaddr (IOMD_IO2CR) - ioaddr (IOMD_BASE);
	    break;
	case 3:
	    io_ctrl = ioaddr (IOMD_IO3CR) - ioaddr (IOMD_BASE);
	    break;
	}
	io_base = ioaddr (IOMD_BASE);

	if (dma[dmanr].invalid) {
	    dma[dmanr].invalid = 0;
	    arm_outb (arm_inb (io_base + io_ctrl) | DMA_CR_C, io_base + io_ctrl);
	    arch_setup_dma (dmanr);
	}

	arm_outb (arm_inb (io_base + io_ctrl) & ~DMA_CR_E, io_base + io_ctrl);
	enable_irq (dmanr + IRQ_DMA0);
    }
}

void arch_disable_dma (unsigned int dmanr)
{
    if (dma[dmanr].active) {
	unsigned int io_ctrl, io_base;

	dma[dmanr].active = 0;
	disable_irq (dmanr + IRQ_DMA0);

	switch (dmanr) {
	case 0:
	    io_ctrl = ioaddr (IOMD_IO0CR) - ioaddr (IOMD_BASE);
	    break;
	case 1:
	    io_ctrl = ioaddr (IOMD_IO1CR) - ioaddr (IOMD_BASE);
	    break;
	case 2:
	    io_ctrl = ioaddr (IOMD_IO2CR) - ioaddr (IOMD_BASE);
	    break;
	case 3:
	    io_ctrl = ioaddr (IOMD_IO3CR) - ioaddr (IOMD_BASE);
	    break;
	}
	io_base = ioaddr (IOMD_BASE);
	arm_outb (arm_inb (io_base + io_ctrl) & ~DMA_CR_E, io_base + io_ctrl);
    }
}
